'use strict';

(function (OC, window, $) {
    class SnapshotSettingsView extends View {
        constructor(handler) {
            super(handler);

            this._fileTypesGroupsList = new FileTypesGroupsList('#filetypes-groups-table', 'filetypes-groups-', {list: '#filetypes-groups-table-tpl', loading: '#filetypes-loading-table-tpl'});
        }

        _postRenderContent() {
            super._postRenderContent();

            this._updateFileTypesGroupsList();

            this._setToolTips();
            this._bindFormActions();
        }

        _updateFileTypesGroupsList() {
            this._fileTypesGroupsList.preparePopulate();
            this._enableFileTypesGroupsControls(false);

            let self = this;
            this.handler().getFileTypesGroups().done(function (fileTypesGroups) {
                self._fileTypesGroupsList.populate(fileTypesGroups);
                self._enableFileTypesGroupsControls();
            }).fail(function () {
                self._fileTypesGroupsList.populate(null);
                self._enableFileTypesGroupsControls();
            });
        }

        _setToolTips() {
            let toolTip = 'Supported duration qualifiers:\nw=weeks, d=days, h=hours, m=minutes';
            $('#snapshot-interval').prop('title', toolTip);
            $('#snapshot-max-age').prop('title', toolTip);
        }

        _bindFormActions() {
            let self = this;

            $('#filetypes-groups-add-new').click(function (event) {
                event.preventDefault();

                self._fileTypesGroupsList.addNewFileTypesGroup();
            });

            $('#settings-save').click(function (event) {
                event.preventDefault();

                self.infoMessage('Saving settings', 'The settings are being saved...');
                self._enableForm(false);

                let settings = self._getSettings();
                self.handler().save(settings).done(function (resp) {
                    self.successMessage('Settings saved', 'The new settings have been saved.');
                    self._enableForm();
                }).fail(function (reason) {
                    self.errorMessage('Settings error', reason.formatError('The settings could not be saved'));
                    self._enableForm();
                });
            });

            $('#settings-reset').click(function (event) {
                event.preventDefault();

                self._enableForm(false);

                self.handler().getDefaults().done(function (resp) {
                    let settings = resp.data();
                    self.handler().save(settings).done(function (resp) {
                        self._setSettings(settings);
                        self._enableForm();
                    }).fail(self._enableForm);
                }).fail(self._enableForm);
            });
        };

        _getSettings() {
            let chunksize = $('#query-chunksize').val().trim();
            if (chunksize === "") {
                chunksize = 0;
            }

            return {
                snapshot_interval: $('#snapshot-interval').val().trim(),
                snapshot_max_age: $('#snapshot-max-age').val().trim(),
                snapshot_filetypes_filter_closed_accounts: $('#snapshot-filetypes-filter-closed-accounts').prop('checked'),

                snapshot_filetypes: JSON.stringify(this._fileTypesGroupsList.getFileTypesGroups()),

                query_chunksize: chunksize,
            };
        }

        _setSettings(settings) {
            $('#snapshot-interval').val(settings.snapshot_interval);
            $('#snapshot-max-age').val(settings.snapshot_max_age);
            $('#snapshot-filetypes-filter-closed-accounts').prop('checked', settings.snapshot_filetypes_filter_closed_accounts);

            this._fileTypesGroupsList.populate(JSON.parse(settings.snapshot_filetypes));

            $('#query-chunksize').val(settings.query_chunksize);
        }

        _enableForm(enable = true) {
            $("#settings-form :input").prop('disabled', !enable);
            $("#settings-form a").each(function ()  {
                if (enable) {
                    $(this).on('click');
                    $(this).css('pointer-events', 'auto');
                } else {
                    $(this).off('click');
                    $(this).css('pointer-events', 'none');
                }
            });
        }

        _enableFileTypesGroupsControls(enable = true) {
            $('#filetypes-groups-add-new').prop('disabled', !enable);
        }
    }

    $(document).ready(function () {
        let handler = new SettingsHandler(OC.generateUrl('/apps/sccuot_ng/config'));

        let view = new SnapshotSettingsView(handler);
        view.render();
    });
})(OC, window, jQuery);
