'use strict';

class Histogram {
    constructor(binCount, floatValues) {
        this._binCount = binCount;
        this._floatValues = floatValues;

        this._data = null;
    }

    calculate(data, minLimit = NaN, maxLimit = NaN) {
        // It is assumed that data is a sorted array in ascending order
        this._data = data;
        if (!isNaN(minLimit) || !isNaN(maxLimit)) {
            this._data = this._filterData(data, minLimit, maxLimit);
        }
        if (this._data.length === 0) {
            return [];
        }

        // Calculate histogram variables
        let valueRange = this.getValueRange();
        let binCount = this._getBinCount(valueRange);
        let binSize = this._getBinSize(valueRange, binCount);

        // Prepare histogram
        let histogram = [];
        for (let i = 1; i <= binCount; ++i) {
            histogram.push({
                    "threshold": Math.min(valueRange.minValue + (binSize * i), valueRange.maxValue),
                    "count": 0,
                }
            );
        }

        // Count values into bins
        for (let value of this._data) {
            let bin = Math.ceil((value - valueRange.minValue) / binSize) - 1;
            bin = Math.max(0, Math.min(bin, binCount-1));
            histogram[bin].count += 1;
        }

        return histogram;
    }

    getValueRange() {
        let minValue = 0;
        let maxValue = 0;
        if (this._data.length !== 0) {
            minValue = this._data[0];
            maxValue = this._data[this._data.length-1];
        }
        return {
            minValue: minValue,
            maxValue: maxValue
        }
    }

    _getBinCount(valueRange) {
        if (this._floatValues) {
            return Math.max(1, this._binCount);
        } else {
            return Math.max(1, Math.min(valueRange.maxValue - valueRange.minValue, this._binCount));
        }
    }

    _getBinSize(valueRange, binCount) {
        let binSize = (valueRange.maxValue - valueRange.minValue) / binCount;
        if (!this._floatValues) {
            binSize = Math.ceil(binSize);
        }
        if (binSize === 0) {
            binSize = valueRange.maxValue;
        }
        return binSize;
    }

    _filterData(data, minLimit, maxLimit) {
        return data.filter((value) => {
            return !((!isNaN(minLimit) && value < minLimit) || (!isNaN(maxLimit) && value > maxLimit));
        });
    }
}
