'use strict';

class ExtendedChart extends Chart {
    constructor(elementId, config) {
        super(elementId, config);

        this._init(elementId);
    }

    _init(element) {
        // Additional plugins
        Chart.register(ChartDataLabels);
        Chart.register("chartjs-plugin-annotation");

        // Default settings overrides
        Chart.defaults.font.size = 12;

        // Custom tooltip positioner
        Chart.Tooltip.positioners.underCursor = function(chartElements, coordinates) {
            coordinates.x += 2;
            return coordinates;
        };

        // Set the chart height to 75% of the view size
        element.height('75vh');
    }

    static createBarChart(elementId, name, labels, data, isHorizontal = false, optionHandler = null) {
        return this._createChart('bar',
            elementId,
            name,
            labels,
            data,
            isHorizontal,
            null,
            (chartConfig) => {
                chartConfig.options.minBarLength = 20;
            },
            optionHandler
        );
    }

    static createLineChart(elementId, name, labels, data, isHorizontal = false, avgAnnotation = null, optionHandler = null) {
        let annotations = {};
        if (avgAnnotation !== null) {
            let averageAnnotation = this._getAverageAnnotation(isHorizontal, avgAnnotation.unit);
            annotations = {...annotations, averageAnnotation};
        }

        return this._createChart('line',
            elementId,
            name,
            labels,
            data,
            isHorizontal,
            annotations,
            null,
            optionHandler
        );
    }

    static _createChart(type, elementId, name, labels, data, isHorizontal, annotations, configHandler, optionHandler) {
        let chartData = this._getDefaultData(name, labels, data);
        chartData.axis = isHorizontal ? 'y' : 'x';

        let chartConfig = this._getDefaultConfig(type, chartData, isHorizontal, annotations);
        chartConfig.options.indexAxis = chartData.axis;

        if (configHandler !== null) {
            configHandler(chartConfig);
        }

        if (optionHandler !== null) {
            optionHandler(chartConfig.options);
        }

        return new ExtendedChart($(elementId), chartConfig);
    }

    static _getDefaultColors() {
        return {
            fill: 'rgba(0, 70, 140, 0.2)',
            border: 'rgb(0, 70, 140)',
            avgAnnotation: 'rgb(255, 0, 0, 0.6)',
            avgAnnotationText: 'white'
        };
    }

    static _getDefaultData(name, labels, data) {
        let colors = this._getDefaultColors();
        return {
            labels: labels,
            datasets: [{
                label: name,
                backgroundColor: colors.fill,
                borderColor: colors.border,
                borderWidth: 2,
                hoverBorderWidth: 3,
                fill: false,
                pointStyle: 'circle',
                pointRadius: 10,
                pointHoverRadius: 13,
                data: data,
            }]
        };
    }

    static _getDefaultConfig(type, data, isHorizontal, annotations) {
        let colors = this._getDefaultColors();
        let options = {
            animation: false,
            maintainAspectRatio: false,
            scales: {
                x: {ticks: {font: {family: "'Open Sans', Frutiger, Calibri, 'Myriad Pro', Myriad, sans-serif", weight: 'bold'}}},
                y: {ticks: {font: {family: "'Open Sans', Frutiger, Calibri, 'Myriad Pro', Myriad, sans-serif", weight: 'bold'}}, min: 0}
            },
            plugins: {
                legend: {display: false},
                tooltip: {
                    mode: 'point',
                    position: 'underCursor',
                    xAlign: 'left',
                    yAlign: 'center',
                    footerColor: '#BBB',
                    footerFont: {weight: 'normal'},
                },
                datalabels: {
                    align: function (context) {
                        let maxValue = isHorizontal ? context.chart.scales.x.max : context.chart.scales.y.max;
                        let value = context.dataset.data[context.dataIndex];
                        if (type === 'line') {
                            return (value < 0.25*maxValue) ? 'end' : 'start';
                        } else {
                            return (value > 0.75*maxValue) ? 'start' : 'end';
                        }
                    },
                    anchor: type === 'line' ? 'center' : 'end',
                    offset: type === 'line' ? 20 : 5,
                    color: 'white',
                    padding: 4,
                    backgroundColor: colors.border,
                    borderColor: colors.border,
                    borderRadius: 3,
                    borderWidth: 2,
                    opacity: 0.85,
                    font: {weight: '500'},
                },
                annotation: {
                    annotations: annotations
                }
            }
        };

        return {
            type: type,
            data: data,
            plugins: [ChartDataLabels, "chartjs-plugin-annotation"],
            options: options,
        };
    }

    static _getAverageAnnotation(isHorizontal, unit) {
        function average(ctx) {
            const values = ctx.chart.data.datasets[0].data;
            return values.reduce((a, b) => a + b, 0) / values.length;
        }

        let colors = this._getDefaultColors();
        return {
            type: 'line',
            borderColor: colors.avgAnnotation,
            borderDash: [6, 6],
            borderDashOffset: 0,
            borderWidth: 1.5,
            label: {
                display: true,
                font: {size: '11', weight: 'normal'},
                position: 'center',
                color: colors.avgAnnotationText,
                backgroundColor: colors.avgAnnotation,
                padding: 4,
                content: (ctx) => 'Average: ' + average(ctx).toFixed(3) + (unit !== null ? ' ' + unit : '')
            },
            scaleID: isHorizontal ? 'x' : 'y',
            value: (ctx) => average(ctx)
        };
    }
}
