'use strict';

const __elementChildGroupName = "group-name";
const __elementChildFileTypes = "file-types";
const __elementChildCollect = "collect";
const __elementChildEdit = "edit";
const __elementChildEditApply = "edit-apply";
const __elementChildEditCancel = "edit-cancel";
const __elementChildDelete = "delete";

const __newEntryPlaceholderName = '_~_newEntryPlaceholder_~_';

class FileTypesGroupsList {
    constructor(elementId, elementChildIdPrefix, templates) {
        this._elementId = elementId;
        this._elementChildIdPrefix = elementChildIdPrefix;
        this._templates = templates;

        this._fileTypesGroups = null;
    }

    preparePopulate() {
        this._fileTypesGroups = null;
        this._renderFileTypesGroupsList(this._templates.loading, false);
    }

    populate(fileTypesGroups) {
        this._fileTypesGroups = new FileTypesGroupsListItems(fileTypesGroups, this._elementChildIdPrefix);
        this._populateGroups();
    }

    addNewFileTypesGroup() {
        this._fileTypesGroups.removePendingNewItems();

        let item = this._fileTypesGroups.newItem(__newEntryPlaceholderName);
        let editItems = this._editGroup(item);

        // "Hide" the internal group name
        editItems[__elementChildGroupName].val("");
    }

    getFileTypesGroups(ordered = false) {
        return this._fileTypesGroups.getFileTypesGroups(ordered);
    }

    _populateGroups() {
        this._renderFileTypesGroupsList(this._templates.list, true, {groups: this._fileTypesGroups.getItems(true)});
        this._bindEditingKeys();
    }

    _renderFileTypesGroupsList(content, enabled, data = {}) {
        this._element().prop('disabled', !enabled);

        renderHandlebarsTemplate(content, this._elementId, data);
        this._bindItemActions();
    }

    _bindEditingKeys() {
        let items = this._fileTypesGroups.getItems();
        for (let groupName in items) {
            let group = items[groupName];
            if (!group.properties.is_editing) {
                continue;
            }

            let editItems = this._fetchEditingItemsFromIndex(group.properties.index);
            if (editItems === null) {
                continue;
            }

            // Bind keyboard events
            let keyboardCallback = (event) => {
                if (event.key === 'Enter' || event.key === 'Escape') {
                    event.preventDefault();

                    if (event.key === 'Enter') {
                        editItems[__elementChildEditApply].click();
                    } else if (event.key === 'Escape') {
                        editItems[__elementChildEditCancel].click();
                    }
                }
            };
            editItems[__elementChildGroupName].keydown(keyboardCallback);
            editItems[__elementChildFileTypes].keydown(keyboardCallback);
        }
    }

    _forEachFileTypesGroup(callback) {
        if (this._fileTypesGroups == null) {
            return;
        }

        let items = this._fileTypesGroups.getItems();
        for (let groupName in items) {
            let group = items[groupName];
            let element = this._element().find('#' + group.properties.index);
            if (element !== null) {
                callback(group, element);
            }
        }
    }

    _bindItemActions() {
        let self = this;
        this._forEachFileTypesGroup(function (group, element) {
            let binder = function (id, callback) {
                let elem = $(element).find('#' + self._getChildElementName(id));
                if (elem !== null) {
                    callback(elem, group);
                }
            }

            binder(__elementChildCollect, function (elem, group) { self._bindItemToggleCollectAction(elem, group); });
            binder(__elementChildEdit, function (elem, group) { self._bindItemEditAction(elem, group); });
            binder(__elementChildEditApply, function (elem, group) { self._bindItemEditApplyAction(elem, group); });
            binder(__elementChildEditCancel, function (elem, group) { self._bindItemEditCancelAction(elem, group); });
            binder(__elementChildDelete, function (elem, group) { self._bindItemDeleteAction(elem, group); });
        });
    }

    _bindItemToggleCollectAction(element, group) {
        element.change(function () {
            group.group.collect = element.prop('checked');
        });
    }

    _bindItemEditAction(element, group) {
        let self = this;
        element.click(function (event) {
            event.preventDefault();

            self._fileTypesGroups.removePendingNewItems();
            self._editGroup(group);
        });
    }

    _bindItemEditApplyAction(element, group) {
        let self = this;
        element.click(function (event) {
            event.preventDefault();

            let editItems = self._fetchEditingItemsFromIndex(group.properties.index);
            if (editItems === null) {
                return;
            }

            if (!self._verifyGroupSettings(group, editItems)) {
                return;
            }

            // Re-add and update entry
            let newGroupName = editItems[__elementChildGroupName].val().trim();
            let newFileTypes = self._parseFileTypesString(editItems[__elementChildFileTypes].val().trim());
            self._fileTypesGroups.updateItem(group, newGroupName, newFileTypes);
            self._fileTypesGroups.disableEditingForAll();
            self._populateGroups();
        });
    }

    _bindItemEditCancelAction(element, group) {
        let self = this;
        element.click(function (event) {
            event.preventDefault();

            self._fileTypesGroups.removePendingNewItems();
            self._fileTypesGroups.disableEditingForAll();
            self._populateGroups();
        });
    }

    _bindItemDeleteAction(element, group) {
        let self = this;
        element.click(function (event) {
            event.preventDefault();

            self._fileTypesGroups.deleteItem(group);
            self._populateGroups();
        });
    }

    _editGroup(group) {
        this._fileTypesGroups.disableEditingForAll();

        group.properties.is_editing = true;
        this._populateGroups();

        // Select the group name of the item being edited
        let editItems = this._fetchEditingItemsFromIndex(group.properties.index);
        if (editItems !== null) {
            editItems[__elementChildGroupName].focus();
            editItems[__elementChildGroupName].select();
        }
        return editItems;
    }

    _verifyGroupSettings(group, editItems) {
        let showError = (elem) => {
            console.log(elem);
            elem.focus();
            elem.addClass(this._getChildElementName('input-error'));
            let self = this;
            elem.on('input', function () {
                $(this).removeClass(self._getChildElementName('input-error'));
            });
            return false;
        }

        let newGroupName = editItems[__elementChildGroupName].val().trim();
        if (newGroupName === "") {
            return showError(editItems[__elementChildGroupName]);
        } else {
            let existingGroup = this._fileTypesGroups.findGroupByName(newGroupName);
            if (existingGroup !== null && existingGroup.properties.index !== group.properties.index) {
                return showError(editItems[__elementChildGroupName]);
            }
        }

        let newFileTypes = this._parseFileTypesString(editItems[__elementChildFileTypes].val().trim());
        if (newFileTypes.length === 0) {
            return showError(editItems[__elementChildFileTypes]);
        }

        return true;
    }

    _fetchEditingItemsFromIndex(groupIndex) {
        let groupRow = this._element().find('#' + groupIndex);
        if (groupRow === null) {
            return null;
        }
        let items = {};
        items[__elementChildGroupName] = groupRow.find('#' + this._getChildElementName(__elementChildGroupName));
        items[__elementChildFileTypes] = groupRow.find('#' + this._getChildElementName(__elementChildFileTypes));
        items[__elementChildEditApply] = groupRow.find('#' + this._getChildElementName(__elementChildEditApply));
        items[__elementChildEditCancel] = groupRow.find('#' + this._getChildElementName(__elementChildEditCancel));
        return items;
    }

    _parseFileTypesString(fileTypes) {
        fileTypes = fileTypes.replaceAll(/[\s|,]+/g, ';').split(';');
        let fileTypesArr = [];
        for (let i = 0; i < fileTypes.length; ++i) {
            let ft = fileTypes[i].trim();
            if (ft !== "") {
                ft = ft.toLowerCase();
                if (!ft.startsWith('.')) {
                    ft = '.' + ft;
                }
                if (fileTypesArr.find((element) => { return element === ft; }) === undefined) {
                    fileTypesArr.push(ft);
                }
            }
        }
        return fileTypesArr;
    }

    _getChildElementName(childName) {
        return this._elementChildIdPrefix + childName;
    }

    _element() {
        return $(this._elementId);
    }
}
