'use strict';

class HistogramPage extends SubPage {
    constructor(title, templateId, histogramElementId, binCountListElement) {
        super(title, templateId);

        this._histogramChart = null;
        this._histogramElementId = histogramElementId;

        this._binCountList = new ListControl(binCountListElement, 'SccuotNG_stats-' + binCountListElement.substring(1));

        this._data = null;
    }

    _getAdditionalContentData(data) {
        this._data = null;

        if (!data.hasOwnProperty('active_snapshot') || data.active_snapshot === null) {
            return;
        }

        this._data = this._getHistogramData(data);
    }

    _getHistogramData(data) {
        return {
            data: null,
            isFloats: false,
        };
    }

    _postRenderContent() {
        super._postRenderContent();

        this._bindBinCountChange();
        this._binCountList.populate(); // This will also draw the chart
    }

    _bindBinCountChange() {
        let self = this;
        this._binCountList.change(function (value) {
            self._createHistogramChart();
        });
    }

    _createHistogramChart() {
        // Destroy any previously created chart
        if (this._histogramChart !== null) {
            this._histogramChart.destroy();
        }

        // Calculate histogram bins
        let limits = this._getHistogramLimits();
        let histogram = new Histogram(this._binCountList.currentValue(), this._data.isFloats);
        let bins = histogram.calculate(this._data.data, limits.minLimit, limits.maxLimit);

        // Prepare labels and data
        let labels = [];
        let data = [];
        let totalUsersCount = 0;
        for (let i = 0; i < bins.length; ++i) {
            let bin = bins[i];
            labels.push('≤ ' + this._formatValue(bin.threshold));
            data.push(bin.count);
            totalUsersCount += bin.count;
        }

        // Create and configure the chart
        let self = this;
        this._histogramChart = ExtendedChart.createBarChart(this._histogramElementId, this.title(), labels, data, false, function (options) {
            let valueRange = histogram.getValueRange();
            options.scales.x.title = {
                display: true,
                text: 'Between ' + self._formatValue(valueRange.minValue) + ' and ' + self._formatValue(valueRange.maxValue) + ' used' +
                    ' (' + formatPluralValue(totalUsersCount, 'user') + ')'
            };

            options.plugins.tooltip.callbacks = {
                title: function (context) {
                    let title = [];
                    for (let ctx of context) {
                        let thresholdLow = self._formatValue(ctx.dataIndex > 0 ? bins[ctx.dataIndex-1].threshold + self._getMinimumDeltaValue() : valueRange.minValue);
                        let thresholdHigh = self._formatValue(bins[ctx.dataIndex].threshold);
                        title.push('Between ' + thresholdLow + ' and ' + thresholdHigh);
                    }
                    return title;
                },
                label: function (context) {
                    let value = parseInt(context.raw);
                    return self._formatUsersCount(value);
                }
            }

            options.plugins.datalabels.formatter = function(value, context) {
                return self._formatUsersCount(value);
            };

            self._getAdditionalOptions(options);
         });
    }

    _getAdditionalOptions(options) {
    }

    _getMinimumDeltaValue() {
        return 0;
    }

    _getHistogramLimits() {
        return {
            minLimit: NaN,
            maxLimit: NaN
        };
    }

    _formatUsersCount(value) {
        return formatPluralValue(value, 'user');
    }

    _formatValue(value) {
        return String(value);
    }
}
