'use strict';

class FileTypesPage extends SubPage {
    constructor() {
        super('File types', '#stats-file-types-tpl');

        this._fileTypesChart = null;
        this._fileTypesChartType = new ListControl('#filetypes-chart-type', 'SccuotNG_stats-file-types-type');

        this._fileTypesCounts = null;
        this._totalFilesCount = 0;
        this._totalFilesSize = 0;
    }

    _getAdditionalContentData(data) {
        this._fileTypesCounts = null;
        this._totalFilesCount = 0;
        this._totalFilesSize = 0;

        if (!data.hasOwnProperty('active_snapshot') || data.active_snapshot === null) {
            return;
        }

        if (!Array.isArray(data.active_snapshot.values.file_types_counts)) { // If the specified field is an array, it means that it is empty
            this._fileTypesCounts = data.active_snapshot.values.file_types_counts;
            this._totalFilesCount = data.active_snapshot.values.files_count.total - data.active_snapshot.values.files_count.obsolete;
            this._totalFilesSize = data.active_snapshot.values.files_size.total - data.active_snapshot.values.files_size.obsolete;
        }
    }

    _postRenderContent() {
        super._postRenderContent();

        this._updateChartTypesList();
        this._bindChartTypeChange();

        if (this._fileTypesCounts !== null) {
            this._createFileTypesChart();
        }
    }

    _getChartTypeHandlers() {
        let self = this;
        return {
            count: {
                minVersion: 1,
                handler: () => self._createFileTypesCountChart()
            },
            size: {
                minVersion: 2,
                handler: () => self._createFileTypesSizesChart()
            }
        };
    }

    _getChartTypeHandler(type) {
        // This should never fail, as the type is always known
        return this._getChartTypeHandlers()[type];
    }

    _createFileTypesChart() {
        // Destroy any previously created chart
        if (this._fileTypesChart !== null) {
            this._fileTypesChart.destroy();
        }

        // Create the chart based on the selected type
        let handler = this._getChartTypeHandler(this._fileTypesChartType.currentValue());
        let labelsCount = handler.handler();

        // The chart description is dynamic, so update it here
        this._updateChartTypeDescription();

        // Update the chart's height
        let chart = $('#filetypes-distribution-chart');
        chart.height(Math.max(labelsCount*60 + 60, chart.height()));
    }

    _createFileTypesCountChart() {
        let self = this;
        return this._createDynamicChart('count', 'Files count', function (options, labels, fileTypes, totalCount) {
            let unlistedCount = Math.max(self._totalFilesCount - totalCount, 0);

            options.scales.x.title = {
                display: true,
                text: formatPluralValue(totalCount, 'file') + ' in ' + formatPluralValue(labels.length, 'group') +
                    ' (' + formatPluralValue(unlistedCount, 'file') + ' unlisted)'
            };

            options.plugins.tooltip.callbacks = {
                footer: function (context) {
                    let footer = [];
                    for (let ctx of context) {
                        footer.push(fileTypes[ctx.dataIndex]);
                    }
                    return footer;
                }
            }

            options.plugins.datalabels.formatter = function(value, context) {
                if (totalCount > 0) {
                    return value.toLocaleString() + ' (' + ((value / totalCount) * 100.0).toFixed(2) + '%)';
                } else {
                    return value;
                }
            };
        });
    }

    _createFileTypesSizesChart() {
        let self = this;
        return this._createDynamicChart('size', 'Files size', function (options, labels, fileTypes, totalSize) {
            let formatValue = function (value) {
                return formatFileSize(value, self._fileSizeUnitFactor, self._fileSizeUnit);
            }

            options.scales.x.title = {
                display: true,
                text: formatValue(totalSize) + ' in ' + formatPluralValue(labels.length, 'group') +
                    ' (' + formatValue(self._totalFilesSize - totalSize) + ' unlisted)',
            };

            options.plugins.tooltip.callbacks = {
                label: function (context) {
                    let value = parseInt(context.raw);
                    return 'Files size: ' + formatValue(value);
                },
                footer: function (context) {
                    let footer = [];
                    for (let ctx of context) {
                        footer.push(fileTypes[ctx.dataIndex]);
                    }
                    return footer;
                }
            }

            options.plugins.datalabels.formatter = function(value, context) {
                let str = formatValue(value);
                if (totalSize > 0) {
                    return str + ' (' + ((value / totalSize) * 100.0).toFixed(2) + '%)';
                } else {
                    return str;
                }
            };
        });
    }

    _createDynamicChart(type, valueName, optionHandler) {
        // Sort descending by type
        let keys = Object.keys(this._fileTypesCounts);
        let self = this;
        keys.sort(function (a, b) { return self._fileTypesCounts[b][type] - self._fileTypesCounts[a][type]; })

        // Prepare labels and data
        let labels = [], fileTypes = [];
        let data = [];
        let totalValue = 0;
        for (let groupName of keys) {
            let group = this._fileTypesCounts[groupName];
            labels.push(groupName);
            fileTypes.push(group.file_types.join(' '));
            data.push(group[type]);
            totalValue += group[type];
        }

        // Create and configure the chart
        this._fileTypesChart = ExtendedChart.createBarChart('#filetypes-distribution-chart', valueName, labels, data, true, function (options) { optionHandler(options, labels, fileTypes, totalValue); });

        return labels.length;
    }

    _updateChartTypesList() {
        // Bind actions, restore saved value, etc.
        this._fileTypesChartType.populate();

        // Hide all types unsupported in the version of the current snapshot
        let self = this;
        this._fileTypesChartType.element().children('option').each(function () {
            let handler = self._getChartTypeHandler($(this).val());
            let isAvailable = self._snapshotVersion >= handler.minVersion;
            $(this).prop('hidden', !isAvailable);
            if ($(this).prop('selected') === true && !isAvailable) {
                self._fileTypesChartType.setCurrentValue('count'); // Count is always available
            }
        });
    }

    _bindChartTypeChange() {
        let self = this;
        this._fileTypesChartType.change(function (value) {
            self._createFileTypesChart();
        });
    }

    _updateChartTypeDescription() {
        let id = '#filetypes-chart-description-' + this._fileTypesChartType.currentValue();
        renderHandlebarsTemplate(id, '#filetypes-chart-description');
    }
}

StatisticsPages.registerPageClass(FileTypesPage);
