'use strict';

(function (OC, window, $) {
    class TimelineView extends View {
        constructor(handler) {
            super(handler);

            this._timelineValues = new TimelineValues();
            this._timelineChart = null;

            this._metricsType = new TimelineValueDropDown(this._timelineValues, '#metrics-list', 'SccuotNG_timeline-active-value');
            this._fileSizeUnits = new FileSizeUnitsDropDown('#filesize-units', 'SccuotNG_timeline-active-unit');
        }

        _postRenderContent() {
            super._postRenderContent();

            this._updateSnapshots();
            this._metricsType.populate();
            this._fileSizeUnits.populate();
        }

        _updateSnapshots() {
            this._showContentTemplate('#timeline-loading-tpl');

            let self = this;
            this.handler().refreshSnapshots().done(function (snapshots) {
                self._bindMetricsTypeChange();
                self._bindFileSizeUnitsChange();
                self._showContentTemplate(snapshots.length === 0 ? '#timeline-empty-tpl' : '#timeline-chart-tpl');
                self._updateTimelineContent();
            }).fail(function () {
                self._showContentTemplate('#timeline-error-tpl');
            });
        }

        _updateTimelineContent() {
            let snapshots = this.handler().snapshots();
            let metricsType = this._metricsType.activeValue();

            // Destroy any previously created chart
            if (this._timelineChart !== null) {
                this._timelineChart.destroy();
            }

            if (snapshots.length > 0 && metricsType !== null) {
                this._createTimelineChart(snapshots, this._timelineValues.value(metricsType));
            }
        }

        _createTimelineChart(snapshots, valueType) {
            // Prepare labels and data
            let labels = [];
            let data = [];
            for (let i = snapshots.length - 1; i >= 0; --i) {
                let snapshot = snapshots[i];
                labels.push(snapshot.timestamp_s.split(' '));
                data.push(this._getSnapshotValue(snapshot, valueType));
            }

            // Create and configure the chart
            let unit = null;
            switch (valueType.getType()) {
                case TimelineValueDescriptor.TYPE_FILESIZE:
                    unit = this._fileSizeUnits.activeUnit().unit;
                    break;
            }
            let self = this;
            this._timelineChart = ExtendedChart.createLineChart('#timeline-chart', valueType.getName(), labels, data, false, {unit: unit}, function (options) {
                options.scales.x.title = {
                    display: true,
                    text: valueType.getName() + ' (' + formatPluralValue(labels.length, 'snapshot') + ')'
                };

                options.plugins.tooltip.callbacks = {
                    title: function (context) {
                        let title = [];
                        for (let ctx of context) {
                            title.push(ctx.label.replace(',', ' '));
                        }
                        return title;
                    },
                    label: function (context) {
                        return self._formatValue(context.raw, valueType);
                    }
                }

                options.plugins.datalabels.formatter = function(value, context) {
                    return self._formatValue(value, valueType);
                };
            });
        }

        _bindMetricsTypeChange() {
            let self = this;
            this._metricsType.change(function (value) {
                self._updateTimelineContent();
            });
        }

        _bindFileSizeUnitsChange() {
            let self = this;
            this._fileSizeUnits.change(function (value) {
                self._updateTimelineContent();
            });
        }

        _showContentTemplate(templateId) {
            renderHandlebarsTemplate(templateId, '#timeline-container-content');
        }

        _getSnapshotValue(snapshot, valueType) {
            let value = this._timelineValues.fromSnapshot(snapshot, valueType.getPath());

            switch (valueType.getType()) {
                case TimelineValueDescriptor.TYPE_FILESIZE:
                    let unit = this._fileSizeUnits.activeUnit();
                    return value/unit.factor;

                default:
                    return value;
            }
        }

        _formatValue(value, valueType) {
            switch (valueType.getType()) {
                case TimelineValueDescriptor.TYPE_FILESIZE:
                    let unit = this._fileSizeUnits.activeUnit();
                    // The value has already been divided by the unit factor, so use 1 here
                    return formatFileSize(value, 1, unit.unit, unit.factor !== 1);

                default:
                    return formatReadableNumber(value, 0);
            }
        }
    }

    $(document).ready(function () {
        let handler = new SnapshotsHandler(OC.generateUrl('/apps/sccuot_ng/snapshot'));

        let view = new TimelineView(handler);
        view.render();
    });
})(OC, window, jQuery);
