<?php

namespace OCA\SccuotNG\Db;

use Exception;

use OCP\AppFramework\Db\{
    DoesNotExistException,
    MultipleObjectsReturnedException,
    QBMapper
};
use OCP\DB\QueryBuilder\IQueryBuilder;
use OCP\IDBConnection;

abstract class ExtMapper extends QBMapper {
	protected int $chunkSize = 0;

	private string $rawTableName = "";

	public function __construct(IDBConnection $db, string $tableName, int $chunkSize = 100000) {
		parent::__construct($db, $tableName);

		$this->chunkSize = $chunkSize;

		$this->rawTableName = $tableName;
	}

	protected function getHighestEntityID(string $field) {
		$query = $this->getQueryBuilder();
		$query->select($field)
			->from($this->rawTableName)
			->orderBy($field, 'DESC');
		$query->setMaxResults(1);
		return $this->queryValue($query, $field);
	}

    protected function chunkedQuery(string $idField, callable $callback, &$data) {
		$offset = 0;
		$highestEntityID = $this->getHighestEntityID($idField);
		while ($offset < $highestEntityID) {
			$query = $this->getQueryBuilder();

			// The provided callback performs the actual query; afterwards, advance the offset
			$maxEntityID = call_user_func_array($callback, array($query, $offset, &$data));
			$offset = max($maxEntityID, $this->advanceOffset($offset));
		}
	}

	protected function addChunkCondition(IQueryBuilder $query, string $field, int $offset, bool $firstWhereClause = false) {
		if ($this->chunkSize > 0) {
			$expr = $query->expr()->gt($field, $query->createPositionalParameter($offset, IQueryBuilder::PARAM_INT));
			if ($firstWhereClause) {
				$query->where($expr);
			} else {
				$query->andWhere($expr);
			}
            $expr = $query->expr()->lte($field, $query->createPositionalParameter($offset + $this->chunkSize, IQueryBuilder::PARAM_INT));
            $query->andWhere($expr);
			$query->orderBy($field, 'ASC');
		}
	}

    protected function getQueryBuilder() : IQueryBuilder {
        return $this->db->getQueryBuilder();
    }

    protected function queryValue($query, string $valueName, int $defaultValue = 0) {
        try {
            $row = $this->queryOne($query);
            return intval($row[$valueName]);
        } catch (Exception $e) {
            return $defaultValue;
        }
    }

	protected function queryValues($query, $values, int $defaultValue = 0) {
		$res = [];
		$row = $this->queryOne($query);
		foreach ($values as $valueName) {
			try {
				$res[$valueName] = intval($row[$valueName]);
			} catch (Exception $e) {
				$res[$valueName] = $defaultValue;
			}
		}
		return $res;
	}

    protected function queryRow($query) {
        return $this->queryOne($query);
    }

    protected function queryRows($query) {
        $stmt = $query->execute();
        $rows = [];
        while ($row = $stmt->fetch()) {
            $rows[] = $row;
        }
        $stmt->closeCursor();
        return $rows;
    }

    protected function queryEntity($query) {
        $row = $this->queryOne($query);
        return $this->mapRowToEntity($row);
    }

    protected function queryEntities($query) {
        $stmt = $query->execute();
        $entities = [];
        while ($row = $stmt->fetch()) {
            $entities[] = $this->mapRowToEntity($row);
        }
        $stmt->closeCursor();
        return $entities;
    }

    private function queryOne($query) {
        $stmt = $query->execute();
        $row = $stmt->fetch();

        if ($row === false || $row === null) {
            $stmt->closeCursor();
            throw new DoesNotExistException($query->getSql());
        }
        $row2 = $stmt->fetch();
        $stmt->closeCursor();

        if (!($row2 === false || $row2 === null)) {
            throw new MultipleObjectsReturnedException($query->getSql());
        } else {
            return $row;
        }
    }
	private function advanceOffset(int $offset): int {
		if ($this->chunkSize > 0) {
			$offset += $this->chunkSize;
		} else {
			$offset = PHP_INT_MAX;
		}
		return $offset;
	}
}
