<?php

namespace OCA\SccuotNG\Db\OC;

use OCA\SccuotNG\Db\ExtMapper;

use OCP\DB\QueryBuilder\IQueryBuilder;
use OCP\IDBConnection;

class FilecacheMapper extends ExtMapper {
    const FILECACHE_TABLE = 'filecache';

	private int $accountsState = 0;

    public function __construct(IDBConnection $db, int $chunkSize, int $accountsState = 0) {
        parent::__construct($db, self::FILECACHE_TABLE, $chunkSize);

		$this->accountsState = $accountsState;
    }

	public function setAccountsState(int $state) {
		$this->accountsState = $state;
	}

	public function queryFiletypesCounts() {
		$data = [
			'dirMimetype' => $this->getDirMimeType(),
			'counts' => []
		];
		$this->chunkedQuery('fileid', array($this, 'queryFiletypesCountsCallback'), $data);
		return $data['counts'];
	}

	protected function queryFiletypesCountsCallback(IQueryBuilder $query, $offset, &$data): int {
		$dirMimetype = $data['dirMimetype'];
		$counts = &$data['counts'];

		$innerQuery = $this->getQueryBuilder();
		$innerQuery->selectAlias('fc.name', 'name')
			->selectAlias('fc.fileid', 'fileid')
			->selectAlias('fc.size', 'filesize')
			->from(self::FILECACHE_TABLE, 'fc');
		$this->joinUserAccounts($innerQuery);
		$innerQuery->where($innerQuery->expr()->notLike('fc.path', $innerQuery->createPositionalParameter('thumbnails/%')))
			->andWhere($innerQuery->expr()->notLike('fc.path', $innerQuery->createPositionalParameter('%_trashbin/%')))
			->andWhere($innerQuery->expr()->like('fc.path', $innerQuery->createPositionalParameter('_%')));
		$this->addDirMimeTypeCondition($innerQuery, $dirMimetype);
		$this->addUserAccountsCondition($innerQuery);
		$this->addChunkCondition($innerQuery, 'fc.fileid', $offset);

		// Outer counting query: Count file types
		$query->automaticTablePrefix(false);
		$query->selectAlias($query->createFunction('COUNT(*)'), 'c')
			->selectAlias($query->createFunction("LOWER(SUBSTRING_INDEX(name,'.',-1))"), 'type')
			->selectAlias($query->createFunction('MAX(fileid)'), 'm')
			->selectAlias($query->createFunction('SUM(filesize)'), 's')
			->from($query->createFunction('(' . $innerQuery->getSQL() . ')'), '_t');
		$query->groupBy('type');
		$query->setParameters($innerQuery->getParameters(), $innerQuery->getParameterTypes());

		// Add up counts & sizes and get highest ID
		$highestID = 0;
		foreach ($this->queryRows($query) as $row) {
			$typeName = '.' . $row['type'];
			if (!array_key_exists($typeName, $counts)) {
				$counts[$typeName] = ['count' => 0, 'size' => 0];
			}
			$counts[$typeName]['count'] += intval($row['c']);
			$counts[$typeName]['size'] += intval($row['s']);
			$highestID = max(intval($row['m']), $highestID);
		}

		return $highestID;
	}

    public function queryTotalFilesCount() {
		$data = [
			'dirMimetype' => $this->getDirMimeType(),
			'totalCount' => 0
		];
		$this->chunkedQuery('fileid', array($this, 'queryTotalFilesCountCallback'), $data);
		return $data['totalCount'];
    }

	protected function queryTotalFilesCountCallback(IQueryBuilder $query, $offset, &$data): int {
		$dirMimetype = $data['dirMimetype'];

		$query->selectAlias($query->createFunction('COUNT(*)'), 'c')
			->selectAlias($query->createFunction('MAX(fc.fileid)'), 'm')
			->from(self::FILECACHE_TABLE, 'fc');
		$this->joinUserAccounts($query);
		$query->where($query->expr()->like('fc.path', $query->createPositionalParameter('_%')))
			->andWhere($query->expr()->notLike('fc.path', $query->createPositionalParameter('thumbnails/%')))
			->andWhere($query->expr()->notLike('fc.path', $query->createPositionalParameter('%_trashbin/%')));
		$this->addDirMimeTypeCondition($query, $dirMimetype);
		$this->addUserAccountsCondition($query);
		$this->addChunkCondition($query,'fc.fileid', $offset);

		// Query values and add up files count
		$values = $this->queryValues($query, ['c', 'm']);
		$data['totalCount'] += $values['c'];

		return $values['m'];
	}

	public function queryTotalFilesSize() {
		$data = [
			'totalSize' => 0
		];
		$this->chunkedQuery('fileid', array($this, 'queryTotalFilesSizeCallback'), $data);
		return $data['totalSize'];
	}

    protected function queryTotalFilesSizeCallback(IQueryBuilder $query, $offset, &$data): int {
        $query->selectAlias($query->createFunction('SUM(fc.size)'), 's')
	        ->selectAlias($query->createFunction('MAX(fc.fileid)'), 'm')
		    ->from(self::FILECACHE_TABLE, 'fc');
	    $this->joinUserAccounts($query);
	    $query->where($query->expr()->eq('path', $query->createPositionalParameter('files')));
	    $this->addUserAccountsCondition($query);
	    $this->addChunkCondition($query, 'fc.fileid', $offset);

	    // Query values and add up files size
	    $values = $this->queryValues($query, ['s', 'm']);
	    $data['totalSize'] += $values['s'];

	    return $values['m'];
    }

	private function joinUserAccounts(IQueryBuilder $query) {
		if ($this->accountsState != 0) {
			$query->join('fc', StoragesMapper::STORAGES_TABLE, 's', 's.numeric_id = fc.storage')
				->join('s', PreferencesMapper::PREFERENCES_TABLE, 'p', "CONCAT('home::', p.userid) = s.id");
		}
	}

	private function addUserAccountsCondition(IQueryBuilder $query) {
		if ($this->accountsState != 0) {
			$query->andWhere(
				$query->expr()->andX($query->expr()->eq('configkey',
					$query->createParameter('enabled')),
					$query->expr()->in('configvalue', $query->createParameter('true'))
				));
		}
	}

	private function addDirMimeTypeCondition(IQueryBuilder $query, $dirMimetype) {
		if ($dirMimetype != -1) {
			$query->andWhere($query->expr()->neq('fc.mimetype', $query->createPositionalParameter($dirMimetype, IQueryBuilder::PARAM_INT)));
		} else {
			$query->andWhere($query->expr()->neq('fc.checksum', $query->createPositionalParameter('')));
		}
	}

    private function getDirMimeType() {
        $mimetypes = new MimetypesMapper($this->db);
        return $mimetypes->queryMimetypeID('httpd/unix-directory');
    }
}
