<?php

namespace OCA\SccuotNG\Db\OC;

use OCA\SccuotNG\Db\ExtMapper;
use OCA\SccuotNG\Util\FileSize;

use OCP\IDBConnection;
use OCP\DB\QueryBuilder\IQueryBuilder;


class PreferencesMapper extends ExtMapper {
    const PREFERENCES_TABLE = 'preferences';

    public function __construct(IDBConnection $db, int $chunkSize) {
        parent::__construct($db, self::PREFERENCES_TABLE, $chunkSize);
    }

	public function queryTotalCount(int $state) {
		$data = [
			'state' => $state,
			'count' => 0
		];
		$this->chunkedQuery('userid', array($this, 'queryTotalCountCallback'), $data);
		return $data['count'];
	}

    protected function queryTotalCountCallback(IQueryBuilder $query, $offset, &$data): int {
		$state = $data['state'];

        $query->selectAlias($query->createFunction('COUNT(*)'), 'c')
	        ->selectAlias($query->createFunction('MAX(id)'), 'm')
            ->from(self::PREFERENCES_TABLE)
			->where($query->expr()->eq('configkey', $query->createNamedParameter('enabled')))
			->andWhere($query->expr()->in('configvalue', $query->createParameter($state)));
	    $this->addChunkCondition($query,'userid', $offset);

	    // Query values and add up count
	    $values = $this->queryValues($query, ['c', 'm']);
	    $data['count'] += $values['c'];

	    return $values['m'];
    }

	public function queryActiveCount(int $since) {
		$data = [
			'since' => $since,
			'count' => 0
		];
		$this->chunkedQuery('userid', array($this, 'queryActiveCountCallback'), $data);
		return $data['count'];
	}

    protected function queryActiveCountCallback(IQueryBuilder $query, $offset, &$data): int {
        $since = $data['since'];

        $query->selectAlias($query->createFunction('COUNT(*)'), 'c')
	        ->selectAlias($query->createFunction('MAX(id)'), 'm')
            ->from(self::PREFERENCES_TABLE)
            ->where($query->expr()->gte('configkey',  $query->createNamedParameter('lastLogin')))
			->andWhere($query->expr()->in('configvalue', $query->createParameter($since)));
	    $this->addChunkCondition($query,'userid', $offset);

	    // Query values and add up count
	    $values = $this->queryValues($query, ['c', 'm']);
	    $data['count'] += $values['c'];

	    return $values['m'];
    }

	public function queryUserStorages() {
		$data = [
			'storages' => []
		];
		$this->chunkedQuery('userid', array($this, 'queryUserStoragesCallback'), $data);
		return $data['storages'];
	}

	protected function queryUserStoragesCallback(IQueryBuilder $query, $offset, &$data): int {
		$storages = &$data['storages'];

		$query->selectAlias('fc.size', 's')
			->selectAlias('p.userid', 'i')
			->from(self::PREFERENCES_TABLE, 'p')
			->join('p', StoragesMapper::STORAGES_TABLE, 's', "s.id = CONCAT('home::', p.userid)")
			->join('s', FilecacheMapper::FILECACHE_TABLE, 'fc', 'fc.storage = s.numeric_id')
			->where($query->expr()->orX(
				$query->expr()->andX($query->expr()->eq('configkey',
				 $query->createParameter('enabled')),
				$query->expr()->in('configvalue', $query->createParameter('true'))
				),
				$query->expr()->isNull('configkey')))
			->andWhere($query->expr()->eq('fc.path', $query->createPositionalParameter('files')));
		$this->addChunkCondition($query,'p.userid', $offset);

		// Add storages and get highest ID
		$highestID = 0;
		foreach ($this->queryRows($query) as $row) {
			$storages[] = intval($row['s']);
			$highestID = max(intval($row['i']), $highestID);
		}

		return $highestID;
	}

	public function queryUserQuotas($defaultQuotas) {
		$data = [
			'defaultQuotas' => $defaultQuotas,
			'quotas' => []
		];
		$this->chunkedQuery('id', array($this, 'queryUserQuotasCallback'), $data);
		return $data['quotas'];
	}

	protected function queryUserQuotasCallback(IQueryBuilder $query, $offset, &$data): int {
		$defaultQuotas = $data['defaultQuotas'];
		$quotas = &$data['quotas'];

		$query->select('id', 'user_id', 'quota', 'backend')
			->from(self::PREFERENCES_TABLE)
			->where($query->expr()->orX(
				$query->expr()->andX($query->expr()->eq('configkey',
					$query->createParameter('enabled')),
					$query->expr()->in('configvalue', $query->createParameter('true'))
				),
				$query->expr()->isNull('configkey')))
			->andWhere($query->expr()->isNotNull('quota'))
			->andWhere($query->expr()->neq('quota', $query->createPositionalParameter('')))
			->andWhere($query->expr()->neq($query->createFunction('LOWER(quota)'), $query->createPositionalParameter('none')));
		$this->addChunkCondition($query,'id', $offset);

		// Add quotas and get highest ID
		$highestID = 0;
		foreach ($this->queryRows($query) as $row) {
			$defaultQuota = $this->findDefaultQuota($defaultQuotas, $row['backend']);
			if (($quota = FileSize::ParseQuotaString($row['quota'], $defaultQuota)) !== false) {
				$quotas[$row['user_id']] = $quota;
			}
			$highestID = max(intval($row['id']), $highestID);
		}

		return $highestID;
	}

	private function findDefaultQuota($defaultQuotas, $backend) {
		foreach ($defaultQuotas as $key => $value) {
			$backendStart = substr($backend, 0, strlen($key));
			if (strcasecmp($key, $backendStart) == 0) {
				return $value;
			}
		}
		return false;
	}
}
