<?php

namespace OCA\SccuotNG\Db;

use JsonSerializable;
use DateTime;

use OCA\SccuotNG\Stats\Value\{IValue,
    FilesCountValue,
	GroupsCountValue,
    FileTypesCountsValue,
    FilesSizeValue,
    SharesCountValue,
    UsersCountValue,
    UsersStorageValue,
	UsersQuotaValue
};

use OCP\AppFramework\Db\Entity;

class Snapshot extends Entity implements JsonSerializable {
	// Versions
	const SNAPSHOT_VERSION_1 = 1;
	const SNAPSHOT_VERSION_2 = 2;
	const SNAPSHOT_VERSION_3 = 3;
	const SNAPSHOT_VERSION_LATEST = Snapshot::SNAPSHOT_VERSION_3;

	const SNAPSHOT_VERSION = Snapshot::SNAPSHOT_VERSION_LATEST;

    // General fields
    const ID_FIELD = 'id';
    const TIMESTAMP_FIELD = 'timestamp';
	const VERSION_FIELD = 'version';
    const USERNAME_FIELD = 'user_name';
    const COMMENT_FIELD = 'comment';

    // Value fields
    const USERS_COUNT_FIELD = 'users_count';
	const GROUPS_COUNT_FIELD = 'groups_count';
    const FILES_COUNT_FIELD = 'files_count';
    const FILES_SIZE_FIELD = 'files_size';
    const SHARES_COUNT_FIELD = 'shares_count';
    const FILE_TYPES_COUNTS_FIELD = 'file_types_counts';
    const USERS_STORAGE_FIELD = 'users_storage';
	const USERS_QUOTA_FIELD = 'users_quota';

    // General values
    protected $timestamp;
	protected $version;
	protected $userName;
    protected $comment;

    // Statistic values
    protected $usersCount;
	protected $groupsCount;
    protected $filesCount;
    protected $filesSize;
    protected $sharesCount;
    protected $fileTypesCounts;
    protected $usersStorage;
	protected $usersQuota;

    public function __construct() {
        $this->createSnapshotValues();

        $this->addType(self::TIMESTAMP_FIELD, 'integer');
	    $this->addType(self::VERSION_FIELD, 'integer');
		$this->addType(self::USERNAME_FIELD, 'string');
        $this->addType(self::COMMENT_FIELD, 'string');

        foreach ($this->values() as $k => $v) {
            $this->addType($k, 'string');
        }
    }

    private function createSnapshotValues() {
        $this->usersCount = new UsersCountValue();
		$this->groupsCount = new GroupsCountValue();
        $this->filesCount = new FilesCountValue();
        $this->filesSize = new FilesSizeValue();
        $this->sharesCount = new SharesCountValue();
        $this->fileTypesCounts = new FileTypesCountsValue();
        $this->usersStorage = new UsersStorageValue();
	    $this->usersQuota = new UsersQuotaValue();
    }

	public function newSnapshot($currentUser, string $comment = '') {
		$this->setTimestamp(time());
		$this->setVersion(self::SNAPSHOT_VERSION);
		$this->setUserName($currentUser != null ? $currentUser->getUserName() : "");
		$this->setComment($comment);
	}

    public function values($fieldNameAsKey = true) {
        if ($fieldNameAsKey) {
            return [
                self::USERS_COUNT_FIELD => $this->usersCount,
	            self::GROUPS_COUNT_FIELD => $this->groupsCount,
                self::FILES_COUNT_FIELD => $this->filesCount,
                self::FILES_SIZE_FIELD => $this->filesSize,
                self::SHARES_COUNT_FIELD => $this->sharesCount,
                self::FILE_TYPES_COUNTS_FIELD => $this->fileTypesCounts,
                self::USERS_STORAGE_FIELD => $this->usersStorage,
	            self::USERS_QUOTA_FIELD => $this->usersQuota,
            ];
        } else {
            return array_filter(get_object_vars($this), function ($varName) { return $this->$varName instanceof IValue; }, ARRAY_FILTER_USE_KEY);
        }
    }

    public function getFormattedTimestamp() {
        $tz = \OC::$server->query('DateTimeZone');
        $date = new DateTime('now', $tz->getTimeZone());
        $date->setTimestamp($this->timestamp);
        return $date->format('Y-m-d H:i:s');
    }

    public function jsonSerialize() {
        $values = [];
        foreach ($this->values() as $k => $v) {
            $values[$k] = $v->jsonSerialize();
        }

        return [
            'id' => $this->id,
            'timestamp' => $this->timestamp,
            'timestamp_s' => $this->getFormattedTimestamp(),
	        'version' => $this->version,
	        'username' => $this->userName,
            'comment' => $this->comment,
            'values' => $values,
        ];
    }

	public function qualifies($minVersion) {
		return $this->version >= $minVersion;
	}

    protected function setter($name, $args) {
        // IValue needs special treatment
        if (property_exists($this, $name) && $this->$name instanceof IValue) {
            $this->$name->setData($args[0]);
            $this->markFieldUpdated($name);
        } else {
            parent::setter($name, $args);
        }
    }

    protected function getter($name) {
        // IValue needs special treatment
        if (property_exists($this, $name) && $this->$name instanceof IValue) {
            return json_encode($this->$name->jsonSerialize());
        } else {
            return parent::getter($name);
        }
    }
}
