<?php

namespace OCA\SccuotNG\Service;

use Exception;
use JsonSerializable;

use OCA\SccuotNG\Data\FileTypesGroups;
use OCP\IConfig;

class ConfigService implements JsonSerializable {
    const SETTING_SNAPSHOT_INTERVAL = 'snapshot_interval';
    const SETTING_SNAPSHOT_MAX_AGE = 'snapshot_max_age';
    const SETTING_SNAPSHOT_FILETYPES = 'snapshot_filetypes';
	const SETTING_SNAPSHOT_FILETYPES_FILTER_CLOSED_ACCOUNTS = 'snapshot_filetypes_filter_closed_accounts';

	const SETTING_QUERY_CHUNKSIZE = 'query_chunksize';

	private string $appName;
    private IConfig $config;

    private array $settings;

    public function __construct($AppName, IConfig $config) {
		$this->appName = $AppName;
        $this->config = $config;
        $this->settings = $this->defaults();

        try {
            $this->load();
        } catch (Exception $e) {
            // Keep defaults if the loaded settings aren't valid
        }
    }

    public function defaults() {
        $settings = [];
        $settings[self::SETTING_SNAPSHOT_INTERVAL] = '1d';
        $settings[self::SETTING_SNAPSHOT_MAX_AGE] = '30d';
        $settings[self::SETTING_SNAPSHOT_FILETYPES] = $this->getDefaultFileTypes();
		$settings[self::SETTING_SNAPSHOT_FILETYPES_FILTER_CLOSED_ACCOUNTS] = false;
		$settings[self::SETTING_QUERY_CHUNKSIZE] = 100000;
        return $settings;
    }

    private function getDefaultFileTypes() {
        $ftg = new FileTypesGroups();
        $ftg->group('Text')->parse('.txt .md .tex');
        $ftg->group('Source code')->parse('.c .cpp .h .hpp .hxx .py .js .php .java .cs .go .html .css');
        $ftg->group('General documents')->parse('.pdf');
        $ftg->group('Office documents')->parse('.doc .docx .odt .xls .xlsx .ods .ppt .pptx .odp');
        $ftg->group('Images')->parse('.jpg .jpeg .png .svg .bmp .gif .tif .tiff .raw');
        $ftg->group('Sounds and Music')->parse('.mp3 .wav .ogg .wma .aiff');
        $ftg->group('Movies')->parse('.avi .mkv .mov .qt .mp4 .flv .mpeg .mpg .ead .xmp');
        $ftg->group('Data')->parse('.mdb .sql .xml .json');
        return json_encode($ftg);
    }

    private function save() {
        foreach ($this->getKeys() as $key) {
            $this->setValue($key, $this->settings[$key]);
        }
    }

    private function load() {
        $settings = [];
        foreach ($this->getKeys() as $key) {
            $settings[$key] = $this->getValue($key, $this->settings[$key]);
        }
        $this->validate($settings);
        $this->settings = $settings;
    }

    private function getKeys() {
        return [
            self::SETTING_SNAPSHOT_INTERVAL,
            self::SETTING_SNAPSHOT_MAX_AGE,
            self::SETTING_SNAPSHOT_FILETYPES,
	        self::SETTING_SNAPSHOT_FILETYPES_FILTER_CLOSED_ACCOUNTS,
	        self::SETTING_QUERY_CHUNKSIZE,
        ];
    }

    public function getSnapshotInterval() {
        return $this->settings[self::SETTING_SNAPSHOT_INTERVAL];
    }

    public function getSnapshotMaxAge() {
        return $this->settings[self::SETTING_SNAPSHOT_MAX_AGE];
    }

    public function getSnapshotFileTypes() {
        $ftg = new FileTypesGroups();

        $setting = $this->settings[self::SETTING_SNAPSHOT_FILETYPES];
        if ($setting != '') {
            $data = json_decode($setting, true);
            $ftg->jsonDeserialize($data);
        }

        return $ftg;
    }

	public function getSnapshotFiletypesFilterClosedAccounts() {
		return $this->settings[self::SETTING_SNAPSHOT_FILETYPES_FILTER_CLOSED_ACCOUNTS];
	}

	public function getQueryChunksize() {
		return $this->settings[self::SETTING_QUERY_CHUNKSIZE];
	}

	public function config() : IConfig {
		return $this->config;
	}

    public function jsonSerialize() {
        return $this->settings;
    }

    public function jsonDeserialize($data) {
        if (!is_array($data)) {
            return;
        }
        $data = array_map('trim', $data);

        $this->validate($data);

        foreach ($this->getKeys() as $key) {
            if (array_key_exists($key, $data)) {
                $this->settings[$key] = $data[$key];
            }
        }

        $this->save();
    }

    private function validate($settings) {
        $validate = function ($settings, string $setting, callable $validator, string $errMsg) {
            if (array_key_exists($setting, $settings)) {
                try {
                    call_user_func($validator, $settings[$setting]);
                } catch (Exception $e) {
                    throw new Exception("{$errMsg}: {$e->getMessage()}");
                }
            }
        };

        $validate($settings, self::SETTING_SNAPSHOT_INTERVAL, 'OCA\\SccuotNG\\Misc\\TimeInterval::parse', 'Snapshot interval invalid');
        $validate($settings, self::SETTING_SNAPSHOT_MAX_AGE, 'OCA\\SccuotNG\\Misc\\TimeInterval::parse', 'Snapshot max. age invalid');
    }

    private function setValue($name, $value) {
        $this->config->setAppValue($this->appName, $name, strval($value));
    }

    private function getValue($name, $default = '') {
        return $this->config->getAppValue($this->appName, $name, strval($default));
    }
}
