<?php

namespace OCA\SccuotNG\Service;

use OCA\SccuotNG\Db\Snapshot;
use OCA\SccuotNG\Db\SnapshotMapper;
use OCA\SccuotNG\Exception\SnapshotNotFoundException;
use OCA\SccuotNG\Exception\CollectorException;

use Exception;

use OCP\AppFramework\Db\{
    DoesNotExistException,
    MultipleObjectsReturnedException
};
use OCP\IDBConnection;

class SnapshotService {
    private IDBConnection $db;

    private SccuotService $sccuot;
	private MetricsService $metrics;
    private ConfigService $config;
    private SnapshotMapper $mapper;

    public function __construct(IDBConnection $db, SccuotService $sccuot, MetricsService $metrics, ConfigService $config, SnapshotMapper $mapper) {
        $this->db = $db;

        $this->sccuot = $sccuot;
		$this->metrics = $metrics;
        $this->config = $config;
        $this->mapper = $mapper;
    }

    public function take(int $chunkSize = 0, string $comment = "") {
        $snapshot = new Snapshot();

        // Fill general snapshot data
	    $snapshot->newSnapshot($this->sccuot->getLoggedInUser(), $comment);

        // Collect all statistic values
        $this->collectSnapshotStats($snapshot, $chunkSize);

        return $this->store($snapshot);
    }

	public function expose(Snapshot $snapshot = null) {
		if ($snapshot == null) {
			$snapshot = $this->getNewest();
		}
		if ($snapshot != null) {
			$this->exposeSnapshotMetrics($snapshot);
		}
	}

    public function getAll() {
        return $this->mapper->getAll();
    }

    public function getNewest() {
        return $this->mapper->getNewest();
    }

    public function find($id) {
        try {
            return $this->mapper->find($id);
        } catch (Exception $e) {
            $this->handleFindException($e);
        }
    }

    public function findByTimestamp($timestamp) {
        try {
            return $this->mapper->findByTimestamp($timestamp);
        } catch (Exception $e) {
            $this->handleFindException($e);
        }
    }

    public function delete(Snapshot $snapshot) {
        $snapshot = $this->find($snapshot->getId());
        $this->mapper->delete($snapshot);
    }

    public function deleteMultiple($ids) {
        return $this->mapper->deleteMultiple($ids);
    }

    public function deleteBefore($timestamp) {
        return $this->mapper->deleteBefore($timestamp);
    }

    private function collectSnapshotStats(Snapshot $snapshot, int $chunkSize) {
        foreach ($snapshot->values(false) as $k => $v) {
            if (($collector = $v->createCollector()) != null) {
                try {
                    $data = $collector->collect($this->db, $snapshot->getTimestamp(), $chunkSize, $this->config);
                    $setter = 'set' . ucfirst($k);
                    $snapshot->$setter($data);
                } catch (Exception $e) {
                    throw new CollectorException("An error occurred while collecting the {$collector->dataName()}: {$e->getMessage()}");
                }
            }
        }
    }

	private function exposeSnapshotMetrics(Snapshot $snapshot) {
		foreach ($snapshot->values(false) as $k => $v) {
			try {
				$v->expose($snapshot, $this->metrics, $this->config);
			} catch (Exception $e) {
				throw new CollectorException("An error occurred while exposing metrics: {$e->getMessage()}");
			}
		}
	}

    private function store(Snapshot $snapshot) {
        try {
            $this->find($snapshot->getId());
            $snapshot = $this->mapper->update($snapshot);
        } catch (SnapshotNotFoundException $e) {
            $snapshot = $this->mapper->insert($snapshot);
        }
        return $snapshot;
    }

    private function handleFindException($e) {
        if ($e instanceof DoesNotExistException || $e instanceof MultipleObjectsReturnedException) {
            throw new SnapshotNotFoundException($e->getMessage());
        } else {
            throw $e;
        }
    }
}
